.global _start

_start:
/*
Contect: We run at address 0x40040, with the MMU enabled. The MMU remaps the
RAM (phys addr 0x30000000) to logical address 0, so we are at 0x30040040 in
physical address space now. Wat we want: Disable the MMU while keeping control
of the PC and then jump to U-boot.
*/
	ldr r0,=mmu_trampoline+0x37000000
	ldr r1,=0x4000
	ldr r2,=uboot_start+0x37000000
	ldr r3,=cpuFlushCache_arm926
	bl mmu_trampoline

/*
Lifted from src/wince/asmstuff.S in Haret
License: GPLv2
*/

		.macro	CPWAIT reg
		mrc	p15, 0, \reg, c2, c0, 0
		mov	\reg, \reg
		sub 	pc, pc, #4
		.endm


@ Code that can disable the MMU and jump to a function.
@ In:   r0 = Physical address of this function
@       r1 = Virtual address of MMU in non-cached ram.
@       r2 = Physical address of code to jump to
@       r3 = cpu flush cache function
        .section .text.trampoline
        .global mmu_trampoline, mmu_trampoline_end
mmu_trampoline:
	@ Store previous registers on stack.
        stmdb   sp, {r4-r14}

        @ Write back all cpu caches
        mov     r4, r0
        mov     r5, r1
        mov     r6, r2
        mov     lr, pc
        mov     pc, r3

        @ Turn off mapping for this code
        mov     r1, r4, lsr #20
        mov     r0, r1, asl #20         @ r0 = funcaddr & 0xfff00000
        orr     r0, r0, #0x00000C00
        orr     r0, r0, #0x00000002     @ r0 |= 0xC02
        str     r0, [r5, r1, asl #2]    @ mmu[funcaddr>>20] = r0

        @ Force MMU update and invalidate I cache
        mov	r0, #0
        mcr	p15, 0, r0, c7, c10, 4	@ Drain write buffer
        mcr	p15, 0, r0, c8, c7, 0	@ Invalidate I+D TLB
        mcr	p15, 0, r0, c7, c5, 0	@ Invalidate I cache & BTB
        CPWAIT  r0

        @ Jump to code with shared virtual/physical mapping
        add     pc, r4, #(1f - mmu_trampoline)

1:      @ Code is running at a vm address that is the same as its
        @ physical address.  Now disable the MMU.
        mrc	p15, 0, r0, c1, c0, 0
        bic	r0, r0, #5		@ MMU & Dcache off
        bic	r0, r0, #0x1000         @ Icache off
        mcr	p15, 0, r0, c1, c0, 0	@ disable the MMU & caches
        CPWAIT  r0

        mov	r0, #0
        mcr	p15, 0, r0, c13, c0, 0	@ clear PID
        mcr	p15, 0, r0, c8, c7, 0	@ invalidate I+D TLB
        mcr	p15, 0, r0, c7, c5, 0	@ Invalidate I cache & BTB
        CPWAIT  r0

        @ MMU is now disabled.  Call function.
        mov     pc, r6




@ Flush CPU caches on arm926 - this MUST be called between
@ take_control()/return_control()
@
@ Dirties r0
		.global cpuFlushCache_arm926
cpuFlushCache_arm926:
1:		mrc	p15, 0, r15, c7, c14, 3		@ test,clean,invalidate
		bne	1b

		mov	r0, #0
		mcr	p15, 0, r0, c7, c5, 0		@ invalidate I cache
		mcr	p15, 0, r0, c7, c10, 4		@ drain WB
		mcr	p15, 0, r0, c8, c7, 0		@ drain I+D TLBs
		mov	pc, r14


uboot_start: